# TkPasMan 2.x core

# Check if the password file, it existent, has the right permissions.
# Return 1 if not OK (OK=only readable by owner)
proc Check {} {

  global conf

  return [expr {[file exists $conf(file)] ? \
      [file attributes $conf(file) -permissions] & 077 : 0}]

}


# Open the password file, lock if it is writable, and load it into the global
# array PW
proc Open {} {

  global conf state PW
  
  # pw file exists? no then create with right permissions
  if {![file exists $conf(file)]} {
    set f [open $conf(file) w 0600]
    close $f
  }
  
  # locked? no then do unless file is not writable.
  if {[file exists $conf(file).lock] || ![file writable $conf(file)]} {
    set state(readonly) 1
  } else {
    set state(readonly) 0
    set f [open $conf(file).lock w 0600]
    puts $f [format %10d [pid]]
    close $f
  }
  
  # read it!
  set f [open $conf(file) r]
  array set PW [read $f]
  close $f
  
  PWChanged 0

}


# Write the password file to disk.
proc Save {} {

  global conf PW

  # save using a backup that is deleted when save was succesful.
  # note that the directory must be writable as well.
  
  if {![catch {
    set f [open $conf(file).new w 0600]
    foreach el [array names PW] {
      puts $f [array get PW $el]
    }
    close $f
  }]} {
    file rename -force $conf(file).new $conf(file)
    PWChanged 0
  } else {
    error "Could not save password file ($conf(file))"
  }
}


# This one is called once, when the user changes the PW array.
# When called with argument 0 the PW array is marked unchanged.
# A trace will be set, to monitor the first change after that.
proc PWChanged {yn args} {

  global PW state

  set state(changed) $yn
  trace [string map {
    0 variable
    1 vdelete
  } $yn] PW wu {PWChanged 1}
}


# Closes the password file. Interacts with the user if changed.
# Returns 1 if the user cancels the close dialog.
proc Close {} {

  global state conf

  if {$state(changed)} {
    if {$conf(onquit-save)} Save else {
      switch -- [tk_messageBox \
          -parent . -icon question -title Save? \
          -message {The file has changed. Do you want to save?} \
          -type yesnocancel -default yes] {
        yes Save
        no {}
        cancel {return 1}
      }
    }
  }
  if {!$state(readonly)} {file delete -force $conf(file).lock}
  return 0
}


# Exit tkpasman, if the user didn't cancel the close dialog.
proc Quit {} {

  global conf

  if {![Close]} {
    if {$conf(onquit-save-options)} SaveConfig
    exit
  }
}


# Write the contents of the global conf array to disk.
proc SaveConfig {} {

  global conf ConfigFile Version

  set conf(geometry) "[winfo width .]x[winfo height .]"

  set f [open $ConfigFile w 0600]
  puts $f "# TkPasMan-$Version\narray set conf {"
  foreach el [lsort [array names conf]] {
    puts $f [array get conf $el]
  }
  puts $f "}\n# The_End"
  close $f
}


# A short hand for the status message.
proc Hint args {

  eval statusMessage .status $args

}


# this one should move to cooltk, when that becomes a separate, multilanguage
# user app-framework...;-)
proc Input {title message value} {

  toplevel [set w .inputwindow] -class Input -bd 8 -relief flat
  wm title $w $title
  wm iconname $w "[wm title .]: $title"
  wm transient $w .

  pack [label $w.l -text $message]
  pack [entry $w.e -textvar $w] -pady 8 -fill x -expand 1
  pack [button $w.ok -text Ok -command "destroy $w" -default active] -side left
  pack [button $w.can -text Cancel -command "set $w {}; destroy $w"] -side right

  focus $w.e
  bind $w.e <Return> "destroy $w"
  bind $w.e <Escape> "set $w {}; destroy $w"

  set ::$w $value
  center $w widget .
  grab $w
  tkwait window $w
  set ::$w
}


proc Help {} {

  global Help

  set w .help
  if {![seeToplevel $w]} {
    toplevel $w -class Help
    wm title $w "TkPasMan Help"
    wm iconname $w "TkPasMan Help"
    pack [button $w.b -text Close -command "wm withdraw $w" -default active] -side bottom
    pack [scrollText $w.t -wrap none] -padx 4 -pady 4 -fill both -expand 1
    set t $w.t.t
    bind $w <Key-Escape> "wm withdraw $w"
    bind $w <Key-Up>    "$t yview scroll -1 units"
    bind $w <Key-Down>  "$t yview scroll  1 units"
    bind $w <Key-Prior> "$t yview scroll -1 pages"
    bind $w <Key-Next>  "$t yview scroll  1 pages"
    bind $w <Key-Home>  "$t see 1.0"
    bind $w <Key-End>   "$t see end"
    $t insert end $Help
    $t config -state disabled
    $t see 1.0
  }
}


proc About {} {

  global Version

  set w .about
  if {![seeToplevel $w]} {
    toplevel $w -class About
    wm title $w "About TkPasMan"
    wm iconname $w "About TkPasMan"
    pack [button $w.b -text Close -command "wm withdraw $w" -default active] -side bottom
    pack [text $w.t -width 50 -height 13 -wrap word] -fill both -expand 1 -padx 8 -pady 8
    bind $w <Key-Escape> "wm withdraw $w"
    $w.t tag configure title -spacing1 3m -spacing3 5m
    $w.t tag configure p -spacing3 3m
    $w.t tag configure url -spacing3 3m
    skinConfigureDefaultTextTags $w.t title p url
    $w.t insert end \
      "TkPasMan version $Version\n"			title	\
      "TkPasMan was written by Wilbert Berendsen <"	p	\
      wbsoft@xs4all.nl					url	\
      "> in Tcl/Tk. This program is Free Software,\
      distributed under the GNU General Public\
      License.\nA copy of the license is included\
      in the distribution.\nHomepage of TkPasMan: see "	p	\
      http://www.xs4all.nl/~wbsoft/linux		url	\
      .\n 						p
    $w.t configure -state disabled
    center $w widget
  }
}


# CreateMainGUI
#
# Creates the main gui and menus. Only called once.
#
proc CreateMainGUI {} {

  global state

  wm protocol . WM_DELETE_WINDOW Quit

  # menu structure
  set m [menu .menu -tearoff no]
  $m add cascade -menu $m.file -label &File
  $m add cascade -menu $m.edit -label &Edit
  $m add cascade -menu $m.opts -label &Options
  $m add cascade -menu $m.help -label &Help

  set s [menu $m.file]
  $s add command -label &Save -command Save -accel Ctrl+S
  $s add command -label &Quit -command Quit -accel Ctrl+Q
  
  set s [menu $m.edit]
  $s add command -label {&Add site} -command Add -accel Ctrl+A
  $s add command -label {&Edit description} -command Edit -accel Ctrl+E
  $s add command -label {&Remove site} -command Remove -accel Ctrl+D

  set s [menu $m.opts]
  $s add check -label {&Always save file on quit} -var conf(onquit-save)
  $s add check -label {&Read only} -var conf(readonly) -command ShowLocked
  $s add check -label {&Confirm Remove site} -var conf(confirm-remove-site)
  $s add check -label {A&llow two pastes} -var conf(allow-two-pastes)
  $s add sep
  $s add check -label {Show &Passwords} -var conf(show-passwd) -command ShowPassWords
  $s add check -label {Show &Menubar} -var conf(show-menu) -command ShowMenuBar -accel Ctrl+M
  $s add check -label {Show &Statusbar} -var conf(show-status) -command ShowStatusBar
  $s add check -label {Remember &Window size} -var conf(save-windowsize)
  if {[llength [skinList]]} {
    $s add cascade -label S&kin -menu $s.skin
    set k [menu $s.skin]
    $k add radio -label None -var conf(skin) -value none -command {skinSet none}
    $k add sep
    foreach skin [skinList] {
      $k add radio -label $skin -var conf(skin) -value $skin -command [list skinSet $skin]
    }
  }
  $s add sep
  $s add check -label {Save &Options on Quit} -var conf(onquit-save-options)
  $s add command -label {Save Options &Now} -command SaveConfig

  set s [menu $m.help]
  $s add command -label &Help -command Help -accel F1
  $s add command -label &About... -command About

  menuSetLabelAccel $m .

  # (scrolled) list box
  pack [scrollListbox .l -listvar state(listVar) -exportsel 0 -selectm browse] \
       -side left -fill both -expand 1

  # right fields.
  pack [label .lu -text Username:] [entry .u -textvar Entry(username)] \
       [label .lp -text Password:] [entry .p -textvar Entry(password)] \
       [label .ls -text Selection:] \
       [radiobutton .1 -var Entry(selectm) -val 1 -text {Username only}] \
       [radiobutton .2 -var Entry(selectm) -val 2 -text {Password only}] \
       [radiobutton .3 -var Entry(selectm) -val 3 -text {Username then Password}] \
       [radiobutton .4 -var Entry(selectm) -val 4 -text {Same, with newlines}] \
       -fill x

  # status bar, not now packed, but later.
  statusBar .status -button {quit Quit Quit}
  
  # bindings
  bind . <3> {tk_popup .menu %X %Y}
  bind .l.l <3> {ListboxSelect [%W index @%x,%y]}
  bind .l.l <<ListboxSelect>> {ListboxSelectCallback}

  # default message
  Hint -persistent "Press F1 for Help"
  
  # remind my birthday ;-)
  if {[set age [wilbertsBirthDay]]} { 
    Hint [format "Today it's Wilbert's birthday! (He's now %d)" $age]
  }

  # keep window title up-to-date
  trace variable state(changed) w UpdateWindowTitle

  # update everything in the display
  ShowMenuBar 0
  ShowStatusBar
  ShowPassWords

  UpdateGlobalState
  UpdateWindowTitle
  UpdateListbox
  UpdateSelection
  
  selection handle . SelectionHandler
}


# show array names in listbox
proc UpdateListbox {} {

  global PW state
  set state(listVar) [lsort -dict [array names PW]]

}


# manipulate screen selection without touching X selection
proc ListboxSelect index {
  .l.l selection clear 0 end
  .l.l selection set $index
  .l.l see $index
  .l.l activate $index
  UpdateSelection
}


# is called from <<ListboxSelect>> event.
proc ListboxSelectCallback {} {

  global conf state Entry

  UpdateSelection
  if {[string compare $state(selected) {}] \
    && [string compare $Entry(selectm) {}] } {

    # set selection to X selection.
    set state(currentSelection) $Entry(selectm)
    set state(selectedUsername) $Entry(username)
    set state(selectedPassword) $Entry(password)
    set state(selectionCount) [expr {$conf(allow-two-pastes) ? 2 : 1}]
    # set up a timer for warning when selection is requested too fast.
    # set conf(selection-delay) to the time value in ms that must pass
    # before the first selection request will be performed without warning.
    set state(selectionTimer)\
      [after $conf(selection-delay) {catch {unset state(selectionTimer)}}]
    switch $state(currentSelection) {
      2       { Hint -persistent "Password selected" }
      default { Hint -persistent "Username selected" }
    }
    selection own -command SelectionLose .
  }
}


proc SelectionLose {} {

  Hint -persistent ""
  Hint "Selection lost"
  
}


proc SelectionHandler {from count} {

  global conf state

  # return empty strings while user is in dialog.
  if {[info exists state(inWarning)]} {return {}}

  if {$from == 0} {
    # when an app requests the selection almost immediately, it is possibly
    # not the user, but some app like Downloader or Klipper. The user is
    # warned.
    if {[info exists state(selectionTimer)]} {
      set state(inWarning) 1
      catch { after cancel $state(selectionTimer)
              unset state(selectionTimer) }
      switch [tk_messageBox -icon warning -title Warning!\
          -type yesno -default no -parent .\
          -message "Warning! An application has requested the selection\
           too fast, within $conf(selection-delay) ms. Is pasting allowed?"] {
        no {
          unset state(inWarning)
          selection clear
          return {}
        }
        yes {
          unset state(inWarning)
          if {[string compare . [selection own]]} {
            Hint "Selection lost anyway"
            return {}
          }
        }
      }
    }
    switch $state(currentSelection) {
      1 {
        set state(selectionString) $state(selectedUsername)
        if {[incr state(selectionCount) -1] == 0} {
          selection clear
          Hint -persistent ""
        } else {
          Hint -persistent "Username still selected"
        }
        Hint "Username pasted"
      }
      2 {
        set state(selectionString) $state(selectedPassword)
        if {[incr state(selectionCount) -1] == 0} {
          selection clear
          Hint -persistent ""
        } else {
          Hint -persistent "Password still selected"
        }
        Hint "Password pasted"
      }
      3 {
        set state(selectionString) $state(selectedUsername)
        set state(currentSelection) 2
        Hint -persistent "Username pasted, password now selected"
      }
      4 {
        set state(selectionString) "$state(selectedUsername)\n"
        set state(currentSelection) 11
        Hint -persistent "Username pasted, password now selected"
      }
      11 {
        set state(selectionString) "$state(selectedPassword)\n"
        if {[incr state(selectionCount) -1] == 0} {
          selection clear
          Hint -persistent ""
        } else {
          Hint -persistent "Password still selected"
        }
        Hint "Password pasted"
      }
    }
  }
  return [string range $state(selectionString) $from [expr {$from + $count - 1}]]
}



# updates entries and disable/enable state.
proc UpdateSelection {} {

  global PW state conf Entry
  
  set s [.l.l curselection]
  if {[llength $s]} {
    trace vdelete Entry w TraceEntry
    set state(selected) [lindex $state(listVar) $s]
    foreach {
      Entry(username)  Entry(password)  Entry(selectm)
    } $PW($state(selected)) break
    trace variable Entry w TraceEntry
    set State normal
  } else {
    trace vdelete Entry w TraceEntry
    set state(selected) {}
    foreach {
      Entry(username)  Entry(password)  Entry(selectm)
    } [list {} {} {}] break
    set State disabled
  }

  if {$conf(readonly) || $state(readonly)} {
    set State disabled
    set AddState disabled
  } else {
    set AddState normal
  }

  foreach w {.u .p .1 .2 .3 .4} {
    $w config -state $State
  }
  set e .menu.edit
  set E [string equal tearoff [$e type 0]]
  $e entryconfig $E -state $AddState
  $e entryconfig [incr E] -state $State
  $e entryconfig [incr E] -state $State
}


proc UpdateGlobalState {} {

  global state
  
  set f .menu.file
  set F [string equal tearoff [$f type 0]]
  $f entryconfig $F -state [expr {$state(readonly) ? "disabled" : "normal"}]

}


# Trace changes in Entry
proc TraceEntry args {

  global PW state Entry
  set PW($state(selected)) \
    [list $Entry(username) $Entry(password) $Entry(selectm)]
}


proc UpdateWindowTitle args {

  global state conf Version

  set State {}
  if {$state(readonly)}   { lappend State readonly } else {
    if {$state(changed)}  { lappend State changed  }
    if {$conf(readonly)}  { lappend State locked   }
  }

  set title "TkPasMan-$Version"
  if {[llength $State]} { append title " ([join $State {, }])" }

  wm title . $title
  wm iconname . $title
}


proc ShowMenuBar {{hint 1}} {

  global conf

  if {$conf(show-menu)} {
    . config -menu .menu
  } else {
    . config -menu {}
    if {$hint} {
      Hint "Use right hand mouse button to popup menu"
    }
  }
}


proc ShowStatusBar {} {

  global conf

  if {$conf(show-status)} {
    pack .status -side bottom -fill x -before .l
  } else {
    pack forget .status
  }
}


proc ShowPassWords {} {

  global conf

  if {$conf(show-passwd)} {
    .p configure -show {}
  } else {
    .p configure -show *
  }
}


proc ShowLocked {{hint 1}} {

  global conf state
  
  UpdateWindowTitle
  UpdateSelection
  if {$hint} {
    if {$state(readonly)} {
      Hint "File not writable at all"
    } elseif {$conf(readonly)} {
      Hint "File write protected"
    } else {
      Hint "File is writable"
    }
  }
}


proc Remove {} {

  global PW conf state

  if {$conf(confirm-remove-site) ? [string equal yes \
      [tk_messageBox -title Alert! \
      -message {Do you really want to remove this site?} \
      -type yesnocancel -default cancel -icon warning -parent .]] : 1} {
    if {[string equal . [selection own]]} { selection clear }
    unset PW($state(selected))
    .l.l selection clear 0 end
    UpdateListbox
    UpdateSelection
    Hint "Site removed"
  } else {
    Hint "Remove cancelled"
  }
}


proc Add {} {

  global PW state
  
  set Site [Input {Add site} "Please enter hostname\nor description:" {}]
  if {[string compare $Site {}]} {
    if {[lsearch -exact $state(listVar) $Site] == -1} {
      set PW($Site) [list {} {} 3]
      UpdateListbox
    }
    ListboxSelect [lsearch -exact $state(listVar) $Site]
  }
}


proc Edit {} {

  global PW state
  
  set Site [Input {Edit description} \
    "Please enter new hostname\nor description:" $state(selected)]
  if {![string equal $Site $state(selected)] && ![string equal $Site {}]} {
    if {[lsearch -exact $state(listVar) $Site] == -1} {
      set PW($Site) $PW($state(selected))
      unset PW($state(selected))
      UpdateListbox
      ListboxSelect [lsearch -exact $state(listVar) $Site]
    } else {
      if {[string equal [
          tk_messageBox -title Alert! -message {Description already exists!} \
          -type retrycancel -default retry -parent .] retry]} {
        after idle Edit
      }
    }
  }
}


#Main...
namespace import cooltk::*

if {$conf(save-windowsize)} {
  wm geometry . $conf(geometry)
}

if {"$conf(skin)" == ""} {
  set conf(skin) [expr {[regexp color [winfo visual .]] ? "Default" : "none"}]
}

# set skin reset command (typically insert some defaults in rec. db)
set s {
  option clear
  option add *Label.anchor			w
  option add *Radiobutton.anchor		w
  option add *Checkbutton.anchor		w
  option add *StatusBar*takeFocus		0
  option add *StatusBar.Button.padY		0
  option add *Scrollbar.takeFocus		0
  option add *ScrollListbox.Listbox.width	30
  option add *Input.Entry.width			30
  option add *Input.Label.font		{Times 18}

  option add *About*AllJustify	center
  option add *About*AllFont	{helvetica 12}		widgetDefault
  option add *About*titleFont	{helvetica 18 bold}	widgetDefault
  option add *About*urlUnderline	1		widgetDefault
}
if {[regexp color [winfo visual .]]} {
  append s {
    option add *About*urlForeground blue	widgetDefault
  }
}
skinSetResetCommand $s ; unset s
skinSet $conf(skin)

#permissions on passwdfile still right?
if {[Check]} {
  wm withdraw .
  switch [tk_messageBox -icon error -type okcancel -default ok -title Warning! \
    -message "File permissions of $conf(file) are not right! The file\
              should be only readable by you. Press OK to fix permissions\
              or Cancel to exit for now."] {
    ok {
      after idle wm deiconify .
      file attributes $conf(file) -permissions 0600
    }
    default {
      puts stderr {tkpasman: error: permissions not fixed -- exit for now}
      exit 1
    }
  }
}
Open
CreateMainGUI
